# -*- coding: utf-8 -*-
###########################################################################
# EOLE - 2013-2017
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
###########################################################################
"""Parsing des fichiers de données AAF
 - parse_aaf_eleves
 - parse_aaf_profs
 - parse_aaf_responsables
 - parse_aaf_etab_educnat
"""
import sys
from lxml import etree  # fast and furious xml library
from logging import getLogger
from pyeole.encode import normalize

from .util import MongoDBGenerator, gen_user_displayname, \
                  modif_fields, pwgen, db, \
                  parse_multi, get_new_id, gen_user_login, \
                  format_name
from .miscutil import user_id
from .config import file_format, domainnames, commonnames, profile_type
if file_format == 'AAF GPEI':
    from .translators.aaf_gpei import AttrsEtab, AttrsEleve, AttrsAdministratif, AttrsResponsable
elif file_format == 'AAF ENT2DVA':
    from .translators.aaf_ent2dva import AttrsEtab, AttrsEleve, AttrsAdministratif, AttrsResponsable
else:
    raise Exception('unknown file_format {0}'.format(file_format))

aaflog = getLogger('AAF')

# mode (complet ou mise a jour) pour l'importation
# le mode d'utilisation peut être fixé de l'extérieur du module
maj_mode = False


# ____________________________________________________________
DEBUG_NUM = 100

# ____________________________________________________________
# global (and unique) sql generator object
sqlgen = MongoDBGenerator()


#_______________________________________________________________________________
# sql inserts


def _get_oriname(etab_name, name):
    return u"{}_{}".format(etab_name, name)


def _get_groupname(uai, name):
    groupname = _get_oriname(format_name(normalize(commonnames.get(uai, uai)).lower()), format_name(normalize(name).lower()))
    double = '__' in groupname
    while double:
        groupname = groupname.replace('__', '_')
        double = '__' in groupname
    return groupname


def _parse_etab_values(db,
                       key,
                       userdata,
                       buffers,
                       force_etab=None,
                       niveau=None,
                       ):
    def _parse_etab_value(value):
        splitted = value.split('$')
        if force_etab:
            etab = force_etab
        else:
            etab = splitted[0]
        if splitted[-1]:
            name = splitted[-1]
        else:
            if not splitted[-2]:
                raise Exception('cannot found the name with {}'.format(value))
            name = splitted[-2]
        etablissement_id, uai = _mongo_fetch_value(etab, "etablissement", "ENTStructureJointure", buffers)
        #group_obj = _mongo_fetch_value(group_name, "group", "ENTMatJointure", buffers)
        # group not already exists
        if key in ['ENTEleveClasses', 'ENTAuxEnsClasses', 'ENTAuxEnsMEF', 'ENTAuxEnsClassesPrincipal']:
            group_type = 'classe'
        elif key == 'ENTEleveNivFormation':
            group_type = 'niveau'
        elif key == 'ENTAuxEnsMatiereEnseignEtab':
            group_type = 'matiere'
        elif key in ['ENTEleveGroupes', 'ENTAuxEnsGroupes']:
            group_type = 'option'
        else:
            raise Exception('unknown key: {}'.format(key))
        group_name = _get_groupname(uai, name)
        if group_name in buffers[group_type]:
            return buffers[group_type][group_name]
        # detection des doublons entre plusieurs types
        real_group_name = group_name
        for buffer_type in buffers:
            if buffer_type == group_type:
                continue
            if real_group_name in buffers[buffer_type]:
                real_group_name = _get_groupname(uai, group_type[0] + name)
                if real_group_name in buffers[group_type]:
                    raise Exception('doublon de nom pour un groupe de type {} avec un groupe de type {}: {}'.format(buffer_type, group_type, name))
        group_obj = {'etablissement': etablissement_id,
                     'etab': etab,
                     'etablissement_uai': uai,
                     'name': real_group_name,
                     'key_name': group_name,
                     'oriName': _get_oriname(uai, name),
                     }
        if group_type == 'classe':
            group_obj['niveau'] = niveau
        buffers[group_type][group_name] = group_obj
        return group_obj
    if key in userdata:
        values = userdata[key]
        new_values = []
        if isinstance(values, list):
            new_values = []
            for value in values:
                if value != '':
                    new_values.append(_parse_etab_value(value))
        elif values != '':
            new_values = _parse_etab_value(values)
        userdata[key] = new_values


def _parse_functions(values, buffers):
    def _parse_function(value):
        splitted = value.split('$')
        etab = splitted[0]
        name = splitted[2]
        discipline = splitted[4]
        etablissement_id, uai = _mongo_fetch_value(etab, "etablissement", "ENTStructureJointure", buffers)
        return {'etablissement': etablissement_id,
                'etablissement_uai': uai,
                'function': name,
                'discipline': discipline}
    if values == '':
        return values
    if isinstance(values, list):
        new_values = []
        for value in values:
            new_values.append(_parse_function(value))
    else:
        new_values = _parse_function(values)
    return new_values


def _parse_ens_class_subjects(values, buffers, name='class'):
    def _parse_ens_class_subject(value):
        splitted = value.split('$')
        etab = splitted[0]
        classe = splitted[1]
        subject = splitted[2]
        return {'etablissement': _mongo_fetch_value(etab, "etablissement", "ENTStructureJointure", buffers)[0],
                name: classe,
                'group': _mongo_fetch_value(subject, "group", "ENTMatJointure", buffers),
                'class': classe}
    if values == '':
        return values
    if isinstance(values, list):
        new_values = []
        for value in values:
            new_values.append(_parse_ens_class_subject(value))
    else:
        new_values = _parse_ens_class_subject(values)
    return new_values


def _parse_attrib_split(value, idx):
    if value == '':
        new_value = value
    elif isinstance(value, list):
        new_value = []
        for val in value:
            new_value.append(val.split('$')[idx])
    else:
        new_value = value.split('$')[idx]
    return new_value


def _parse_aaf_profadmin(db, usertype, userdata, buffers):
    if usertype == 'administratif':
        unused_fields = ['ENTAuxEnsClassesPrincipal', 'ENTAuxEnsCategoDiscipline',
                         'ENTAuxEnsDisciplinesPoste', 'ENTAuxEnsGroupes',
                         'ENTAuxEnsGroupesMatieres']
        for field in unused_fields:
            if field in userdata:
                   del(userdata[field])
    _mongo_fetch_values(userdata,
                        "ENTPersonStructRattach",
                        "etablissement",
                        "ENTStructureJointure",
                        buffers,
                        )
    if usertype == 'enseignant':
        for key in ['ENTAuxEnsMatiereEnseignEtab', 'ENTAuxEnsClasses', 'ENTAuxEnsMEF', 'ENTAuxEnsClassesPrincipal', 'ENTAuxEnsGroupes']:
            _parse_etab_values(db,
                               key,
                               userdata,
                               buffers,
                               )
        if 'ENTAuxEnsDisciplinesPoste' in userdata:
            userdata['ENTAuxEnsDisciplinesPoste'] = _parse_attrib_split(userdata['ENTAuxEnsDisciplinesPoste'], 0)
        if 'ENTAuxEnsCategoDiscipline' in userdata:
            userdata['ENTAuxEnsCategoDiscipline'] = _parse_attrib_split(userdata['ENTAuxEnsCategoDiscipline'], 1)
        if u"ENTAuxEnsClassesMatieres" in userdata:
            userdata[u"ENTAuxEnsClassesMatieres"] = _parse_ens_class_subjects(userdata[u"ENTAuxEnsClassesMatieres"], buffers)
        if u"ENTAuxEnsGroupesMatieres" in userdata:
            userdata[u"ENTAuxEnsGroupesMatieres"] = _parse_ens_class_subjects(userdata[u"ENTAuxEnsGroupesMatieres"], buffers, 'group')
    if 'ENTPersonFonctions' in userdata:
        userdata['ENTPersonFonctions'] = _parse_functions(userdata['ENTPersonFonctions'], buffers)
    if sys.version_info[0] < 3:
        userdata['gecos'] = str(gen_user_displayname(userdata['sn'].encode('utf-8'), userdata['givenName'].encode('utf-8')))
    else:
        userdata['gecos'] = str(gen_user_displayname(userdata['sn'], userdata['givenName']))


def delete_profs(ids):
    db_profs = set()
    db_admins = set()
    for prof in db.fetchall('user', attr=u'UserType', value='enseignant', projection={'id':1}):
        db_profs.add(prof['id'])
    for admin in db.fetchall('user', attr=u'UserType', value='administratif', projection={'id':1}):
        db_admins.add(admin['id'])
    users_to_be_deleted = (db_profs | db_admins) - set(ids)
    for user in users_to_be_deleted:
        db_users = db.fetchone('user',
                               "id",
                               user,
                               )
        if user in db_profs:
            usertype = 'enseignant'
        else:
            usertype = 'administratif'
        modif_fields(db_users, sqlgen, 'DELETE',
                     modif_type='DELTADELETE', tablename=usertype, fromxml=True, db=db,
                     no_bulk=True)
    db.commit('user', no_bulk=True)


def _mongo_fetch_value(ori_value,
                       tablename,
                       filterby,
                       buffers,
                       ):
    if ori_value in buffers[tablename]:
        new_value = buffers[tablename][ori_value]
    else:
        fetch_result = db.fetchone(tablename,
                                   filterby,
                                   ori_value,
                                   )
        if fetch_result is not None:
            if tablename == 'etablissement':
                new_value = (fetch_result[u'_id'], fetch_result["ENTStructureUAI"])
            else:
                new_value = fetch_result[u'_id']
        else:
            if tablename == 'etablissement':
                new_value = (ori_value, "-")
            else:
                new_value = ori_value
        buffers[tablename][ori_value] = new_value
    return new_value


def _mongo_fetch_values(userdata,
                        key,
                        tablename,
                        filterby,
                        buffers,
                        ):
    if key not in userdata:
        return
    values = userdata[key]
    if isinstance(values, list) and len(values) == 1:
        values = values[0]
    if values == '':
        return ''
    if isinstance(values, list):
        new_values = []
        for val in values:
            nvalues = _mongo_fetch_value(val,
                                         tablename,
                                         filterby,
                                         buffers,
                                         )
            if tablename == 'etablissement':
                nvalues = nvalues[0]
            new_values.append(nvalues)
    else:
        new_values = _mongo_fetch_value(values,
                                        tablename,
                                        filterby,
                                        buffers,
                                        )
        if tablename == 'etablissement':
            new_values = new_values[0]
    userdata[key] = new_values


def _parse_aaf_student(db,
                       userdata,
                       uid,
                       buffers,
                       ):
    # ajout de l'eleve a la classe et au groupe
    _parse_etab_values(db,
                       'ENTEleveGroupes',
                       userdata,
                       buffers)
    if 'ENTElevePersRelEleve' in userdata:
        # TODO: on avait évoqué la possibilité de pouvoir limiter la création
        # aux responsables "1" & "2" (resp.split('$')[1] in ['1', '2'])
        # ce qui revient à être dans la liste
        # de "ENTEleveAutoriteParentale" (cf. Scribe)
        for key in ('ENTEleveParents', 'ENTElevePere', 'ENTEleveMere', 'ENTEleveAutoriteParentale'):
            _mongo_fetch_values(userdata,
                                key,
                                "user",
                                "ENTPersonJointure",
                                buffers,
                                )
        if isinstance(userdata['ENTElevePersRelEleve'], list):
            resps = userdata['ENTElevePersRelEleve']
        else:
            resps = userdata['ENTElevePersRelEleve'].split('\n')
    else:
        resps = []
    if 'ENTEleveCodeEnseignements' in userdata:
        _mongo_fetch_values(userdata,
                            u'ENTEleveCodeEnseignements',
                            'group',
                            "ENTMatJointure",
                            buffers,
                            )

    _mongo_fetch_values(userdata,
                        "ENTPersonStructRattach",
                        "etablissement",
                        "ENTStructureJointure",
                        buffers,
                        )
    # niveau in same etab than class
    force_etab = userdata['ENTEleveClasses'].split('$')[0]
    _parse_etab_values(db,
                       'ENTEleveNivFormation',
                       userdata,
                       buffers,
                       force_etab=force_etab,
                       )
    _parse_etab_values(db,
                       'ENTEleveClasses',
                       userdata,
                       buffers,
                       niveau=userdata['ENTEleveNivFormation']['name'],
                       )
    if sys.version_info[0] < 3:
        userdata['gecos'] = str(gen_user_displayname(userdata['sn'].encode('utf-8'), userdata['givenName'].encode('utf-8')))
    else:
        userdata['gecos'] = str(gen_user_displayname(userdata['sn'], userdata['givenName']))

    for resp in resps:
        numresp = resp.split('$')[0]
        if numresp in buffers['eleves_responsable']:
            if uid not in buffers['eleves_responsable'][numresp]:
                buffers['eleves_responsable'][numresp].append(uid)
        else:
            buffers['eleves_responsable'][numresp] = [uid]


def _parse_aaf_resp(userdata):
    # ____________________________________________________________
    # remplissage des élèves associé au userdata
    # à partir des inforamtions bufferisées
    if sys.version_info[0] < 3:
        userdata['gecos'] = str(gen_user_displayname(userdata['sn'].encode('utf-8'), userdata['givenName'].encode('utf-8')))
    else:
        userdata['gecos'] = str(gen_user_displayname(userdata['sn'], userdata['givenName']))


def _modif_type_to_action(modif_type):
    if modif_type in ('CREATE', 'DELTACREATE'):
        action = 'ADD'
    elif modif_type in ('UPDATE', 'LDIFEXPORT', 'DELTAUPDATE', 'DELTAMOVE'):
        action = 'UPDATE'
    elif modif_type == 'DELTADELETE':
        action == 'DELETE'
    else:
        raise Exception('Unknown modif_type')
    return action


def gen_profil(smb_server,
               login,
               ):
    """
        Génération du sambaProfilePath selon le profil demandé
    """
    if smb_server is None:
        return smb_server
    if profile_type == 'none':
        return None
    elif profile_type == 'itinerant':
        # profil itinérant
        return "\\\\{0}\\{1}\\profil".format(smb_server, login)
    elif profile_type == 'local':
        # profil local à la station
        return ""
    raise Exception('type de profile "{}" inconnu'.format(profile_type))


def gen_homedirectory(smb_server,
                      username,
                      ):
    return '\\\\{0}\\{1}\\perso'.format(smb_server, username)



# ____________________________________________________________
# Generic parsing function
def parse_aaf(type_, filename, aaftype, modif_type, parse_type, allow_bulk):
    """parsing depuis AAF"""
    if allow_bulk and modif_type == 'CREATE' and not maj_mode:
        #cas aaf-complet (non maj)
        no_bulk = False
    else:
        no_bulk = True
    num = 0
    aaflog.info("Lecture des {}s...".format(type_))
    login_buffer = []
    rnd_buffer = list()
    if parse_type == 'ADD':
        parse_tag = 'addRequest'
        parse_find = 'attributes/attr'

    elif parse_type == 'MODIFY':
        parse_tag = 'modifyRequest'
        parse_find = 'modifications/modification'

    action = _modif_type_to_action(modif_type)

    context = etree.iterparse(filename, events=('end',), tag=parse_tag)
    users_id = []
    buffers = {'etablissement': {},
               'group': {},
               'user': {},
               'eleves_responsable': {}}
    def _set_buffer_from_aggregation(typ, row):
        if isinstance(row, list):
            for r in row:
                _set_buffer_from_aggregation(typ, r)
        elif row is None:
            pass
            #raise Exception('unable to find any "{}", almost one should already imported with student'.format(typ))
        else:
            buffers[typ][row['key_name']] = row
    for typ in ('classe', 'option', 'niveau', 'matiere'):
        buffers[typ] = {}
        for row in db.query_groupname(typ):
            _set_buffer_from_aggregation(typ, row)
    for key in ['ENTAuxEnsClasses', 'ENTAuxEnsMEF', 'ENTAuxEnsClassesPrincipal']:
        for row in db.aggregate('enseignant', '$' + key):
            _set_buffer_from_aggregation('classe', row[u'_id'])
    for row in db.aggregate('enseignant', '$ENTAuxEnsGroupes'):
        _set_buffer_from_aggregation('option', row[u'_id'])
    for event, tdata in context:
        _type = tdata.find('operationalAttributes/attr/value').text
        if _type != aaftype:
            aaflog.info("%s pas un {} :(".format(_type, type_))
            continue
        #**Attention** identifier/id == ENTPersonJointure
        userdata = user_id(tdata)
        users_id.append(userdata['id'])
        if type_ == 'eleve':
            attrs_translators = AttrsEleve
        elif type_ == 'administratif':
            attrs_translators = AttrsAdministratif
        elif type_ == 'responsable':
            attrs_translators = AttrsResponsable
        else:
            raise Exception('unknown type_ for {}'.format(type_))
        for attr in tdata.findall(parse_find):
            balise, vals = parse_multi(attr)
            if balise in attrs_translators:
                balise = attrs_translators[balise]
            userdata[balise] = vals
        if 'ENTPersonJointure' not in userdata:
            userdata['ENTPersonJointure'] = userdata['id']
        retrieved = False
        if 'givenName' in userdata:
            if isinstance(userdata['givenName'], list):
                userdata['givenName'] = [val.capitalize() for val in userdata['givenName']]
            else:
                userdata['givenName'] = userdata['givenName'].capitalize()
        if type_ == 'administratif':
            #Règle de répartition des personnes
            #cf. Annexe 4 (Alimentation depuis le SI MEN) version 4.2
            if isinstance(userdata['ENTPersonFonctions'], list):
                function = userdata['ENTPersonFonctions'][0]
            else:
                function = userdata['ENTPersonFonctions']
            if '$ENS$' in function or '$DOC$' in function:
                usertype = 'enseignant'
            else:
                usertype = 'administratif'
        else:
            usertype = type_
        if maj_mode or parse_type == 'MODIFY':
            db_user = db.fetchone("user",
                                  "id",
                                  userdata['id'],
                                  aaftype=usertype,
                                  )
            if db_user is None:
                modif_type = 'DELTACREATE'
                db_user = {}
                if parse_type == 'MODIFY':
                    aaflog.warn("ATTENTION: l'{1} {0} censé être mis a jour "
                             "n'existe pas dans la base".format(userdata['id'], type_))
                action = 'ADD'
            else:
                current_modif_type = db_user['FieldActionType']
                if current_modif_type in ['CREATE', 'DELTACREATE']:
                    modif_type = 'DELTACREATE'
                elif current_modif_type == 'DELTADELETE':
                    modif_type = 'DELTADELETE'
                retrieved = True
                action = 'MODIFY'
        else:
            db_user = {}
        if 'uid' not in db_user:
            uid = userdata['uid'] = get_new_id(rnd_buffer)
            if retrieved:
                aaflog.error("ATTENTION: l'utilisateur {0} aurait dû avoir un "
                          "uid".format(userdata['id']))
        else:
            uid = userdata['uid'] = db_user['uid']

        if 'userPassword' not in db_user:
            userdata['userPassword'] = pwgen()
            if retrieved:
                aaflog.error("ATTENTION: l'eleve {0} aurait dû avoir un mot de "
                          "passe".format(userdata['id']))
        else:
            userdata['userPassword'] = db_user['userPassword']

        # **IMPORTANT** : this login shall not change once set at the user's creation
        # from the AAF point of vue, the key is the id
        # for samba, it's the generated cn
        if 'ENTPersonLogin' not in db_user:
            if retrieved:
                aaflog.error("ATTENTION: l'eleve {0} aurait dû avoir un "
                          "login".format(userdata['id']))
            userdata['ENTPersonLogin'] = gen_user_login(userdata['sn'],
                                                        userdata['givenName'],
                                                        uid,
                                                        type_,
                                                        userdata['id'],
                                                        login_buffer,
                                                        )
        else:
            userdata['ENTPersonLogin'] = db_user['ENTPersonLogin']

        if 'ENTPersonStructRattach' in userdata:
            etab = _mongo_fetch_value(userdata["ENTPersonStructRattach"],
                                      "etablissement",
                                      "ENTStructureJointure",
                                      buffers,
                                      )[1]
        else:
            etab = None
        if usertype == 'eleve':
            _parse_aaf_student(db,
                               userdata,
                               uid,
                               buffers,
                               )
        elif type_ == 'responsable':
            _parse_aaf_resp(userdata)
        elif type_ == 'administratif':
            _parse_aaf_profadmin(db, usertype, userdata, buffers)
        smb_server = domainnames.get(etab)
        if smb_server:
            smb_server = smb_server.upper()
        if smb_server is None:
            homedirectory = None
        else:
            homedirectory = gen_homedirectory(smb_server,
                                              userdata['ENTPersonLogin'],
                                              )
        profilepath = gen_profil(smb_server,
                                 userdata['ENTPersonLogin'],
                                 )
        launch_modif_fields = True
        userdata['homedirectory'] = homedirectory
        userdata['profilepath'] = profilepath
        if retrieved:
            unlisted_key = set([u'FieldFromXMLImportActionDate', u'_id', u'FieldModifiedStatus', u'FieldActionType', u'UserType', u'FieldToLDIFFExportActionDate'])
            key_old = set(db_user.keys())
            key_new = set(userdata.keys())
            is_modified = False
            if len(key_old - key_new - unlisted_key) != 0 or len(key_new - key_old) != 0:
                is_modified = True
                aaflog.debug("some fields in the database ({}) and the current user ({}) does not match for the user: {}".format(
                          key_old - unlisted_key, key_new, userdata['id']))
            else:
                for key in key_new:
                    if db_user[key] != userdata[key]:
                        is_modified = True
                        aaflog.debug("some entries in the database ({}) and the current user ({}) does not match for the user".format(
                                  db_user[key], userdata[key], userdata['id']))
                        break
            if is_modified:
                if db_user['FieldActionType'] == 'DELTADELETE':
                    pass
                elif db_user['FieldActionType'] in ['CREATE', 'DELTACREATE']:
                    modif_type = 'DELTACREATE'
                else:
                    if db_user.get('ENTPersonStructRattach') != userdata.get('ENTPersonStructRattach'):
                        modif_type = 'DELTAMOVE'
                        if 'ENTPersonStructRattachsOld' not in userdata:
                            userdata['ENTPersonStructRattachsOld'] = []
                        userdata['ENTPersonStructRattachsOld'].append(db_user.get('ENTPersonStructRattach'))
                    else:
                        modif_type = 'DELTAUPDATE'
                action = 'UPDATE'
            else:
                launch_modif_fields = False
        if launch_modif_fields:
            # ____________________________________________________________
            # traitement de l'utilisateur
            modif_fields(userdata, sqlgen, action, modif_type, tablename=usertype,
                        fromxml=True, db=db, no_bulk=no_bulk)
            num += 1
        if num % DEBUG_NUM == 0:
            aaflog.debug("{} {}s lus...".format(num, type_))
        tdata.clear()
        while tdata.getprevious() is not None:
            del tdata.getparent()[0]
    del context
    login_buffer = []

    db.commit("user", no_bulk=no_bulk)
    aaflog.info("sous total : {} {}s".format(num, type_))
    return num, users_id

# ____________________________________________________________
# Extraction AAF Eleves et Responsables
def parse_aaf_eleves(eleve_file, allow_bulk, modif_type='CREATE', parse_type='ADD'):
    """parsing des élèves depuis AAF"""
    return parse_aaf('eleve', eleve_file, 'Eleve', modif_type, parse_type, allow_bulk)


def delete_eleves(ids):
    db_users = set()
    for user in db.fetchall('user', attr=u'UserType', value='eleve', projection={'id':1}):
        db_users.add(user['id'])
    users_to_be_deleted = db_users - set(ids)
    for user in users_to_be_deleted:
        db_user = db.fetchone('user',
                              "id",
                              user,
                              )
        modif_fields(db_user, sqlgen, 'DELETE', modif_type='DELTADELETE', tablename='eleve',
                     fromxml=True, db=db, no_bulk=True)
    db.commit('user', no_bulk=True)


def parse_aaf_responsables(responsables_file, allow_bulk, modif_type='CREATE', parse_type='ADD'):
    """Parsing XML des responsables depuis AAF
    :responsables_file: file name like ``*_PersRelEleve_*.xml``
    """
    return parse_aaf('responsable', responsables_file, 'PersRelEleve', modif_type, parse_type, allow_bulk)


def delete_responsables(ids):
    db_users = set()
    for user in db.fetchall('user', attr=u'UserType', value='responsable', projection={'id':1}):
        db_users.add(user['id'])
    users_to_be_deleted = db_users - set(ids)
    for user in users_to_be_deleted:
        db_user = db.fetchone('user',
                              "id",
                              user,
                              )
        modif_fields(db_user, sqlgen, 'DELETE', modif_type='DELTADELETE', tablename='responsable',
                     fromxml=True, db=db, no_bulk=True)
    db.commit('user', no_bulk=True)


def parse_aaf_profs(profs_file, allow_bulk, modif_type='CREATE', parse_type='ADD'):
    """ parsing des professeurs depuis AAF
    Extraction AAF Professeurs + Administratifs
    Fichier de type : ``ENT_$RNE_Complet_$DATE_PersEducNat_0000.xml``
    """
    return parse_aaf('administratif', profs_file, 'PersEducNat', modif_type, parse_type, allow_bulk)


# ____________________________________________________________
def parse_aaf_etab_educnat(etab_file,
                           allow_bulk,
                           modif_type='CREATE',
                           parse_type='ADD',
                           ):
    """Parsing XML des etablissements depuis AAF
    :etab_file: file name like ``*_EtabEducNat_*.xml``
    """
    # init
    if allow_bulk and modif_type == 'CREATE' and not maj_mode:
        #cas aaf-complet (non maj)
        no_bulk = False
    else:
        no_bulk = True
    num = 0
    action = _modif_type_to_action(modif_type)
    aaflog.info("Lecture des etablissements...")
    etablissements_id = []
    if parse_type == 'ADD':
        parse_tag = 'addRequest'
        parse_find = 'attributes/attr'

    elif parse_type == 'MODIFY':
        parse_tag = 'modifyRequest'
        parse_find = 'modifications/modification'

    context = etree.iterparse(etab_file, events=('end',), tag=parse_tag)
    # ____________________________________________________________
    # dict construction by XML parsing
    for event, tresp in context:
        _type = tresp.find('operationalAttributes/attr/value').text
        if _type != 'EtabEducNat':
            aaflog.info("%s n'est pas un etablissement :(" % _type)
            continue
        etablissement = dict()
        etabid = tresp.find('identifier/id').text
        if sys.version_info[0] < 3:
            etablissement['id'] = unicode(etabid)
        else:
            etablissement['id'] = str(etabid)
        etablissements_id.append(etablissement['id'])
        for attr in tresp.findall(parse_find):
            balise, vals = parse_multi(attr)
            if balise in AttrsEtab:
                balise = AttrsEtab[balise]
            etablissement[balise] = vals
        retrieved = False
        if maj_mode or parse_type == 'MODIFY':
            db_etab = db.fetchone('etablissement',
                                  "id",
                                  etablissement['id'],
                                  )
            if db_etab is None:
                modif_type = 'DELTACREATE'
                if not maj_mode:
                    aaflog.warn("ATTENTION: l'établissement {0} censé être mis a jour "
                             "n'existe pas dans la base".format(etablissement['id']))
                action = 'ADD'
            else:
                current_modif_type = db_etab['FieldActionType']
                if maj_mode:
                    modif_type = 'DELTAUPDATE'
                else:
                    if current_modif_type in ['CREATE', 'DELTACREATE']:
                        modif_type = 'DELTACREATE'
                    elif current_modif_type == 'DELTADELETE':
                        modif_type = 'DELTADELETE'
                retrieved = True
                action = 'UPDATE'
        if 'ENTStructureGroupes' in etablissement and isinstance(etablissement['ENTStructureGroupes'], list):
            new_value = []
            for value in etablissement['ENTStructureGroupes']:
                spl = value.split('$')
                new_value.append({'class': spl[-1], 'group': spl[0]})
            etablissement['ENTStructureGroupes'] = new_value
        if u"ENTEtablissementBassin" in etablissement and '$' in etablissement[u"ENTEtablissementBassin"]:
            etablissement[u"ENTEtablissementBassin"] = etablissement[u"ENTEtablissementBassin"].split('$')[1]
        etablissement[u'etab_name'] = domainnames.get(etablissement[u'ENTStructureUAI'].upper(), etablissement[u'ENTStructureUAI'])
        launch_modif_fields = True
        if retrieved:
            unlisted_key = set([u'FieldFromXMLImportActionDate', u'_id', u'FieldModifiedStatus', u'FieldActionType', u'', u'FieldToLDIFFExportActionDate'])
            key_old = set(db_etab.keys())
            key_new = set(etablissement.keys())
            is_modified = False
            etab_name_has_change = False

            if len(key_old - key_new - unlisted_key) != 0 or len(key_new - key_old) != 0:
                is_modified = True
                aaflog.debug("some fields in the database ({}) and the current etablissement ({}) does not match for the etablissement: {}".format(
                          key_old - unlisted_key, key_new, etablissement['ENTStructureUAI']))
            else:
                for key in key_new:
                    if db_etab[key] != etablissement[key]:
                        is_modified = True
                        if key == u'etab_name':
                            etab_name_has_change = True
                        aaflog.debug("some entries in the database ({}) and the current etablissement ({}) does not match for the etablissement".format(
                                  db_etab[key], etablissement[key], etablissement['ENTStructureUAI']))
                        break
            if is_modified:
                if db_etab[u'FieldActionType'] in ['CREATE', 'DELTACREATE'] or etab_name_has_change:
                    modif_type = 'DELTACREATE'
                else:
                    modif_type = 'DELTAUPDATE'
                action = 'UPDATE'
            else:
                launch_modif_fields = False
        if launch_modif_fields:
            # ____________________________________________________________
            # database insert
            modif_fields(etablissement, sqlgen, action, modif_type=modif_type, tablename="etablissement",
                        fromxml=True, db=db, no_bulk=no_bulk)
            # ____________________________________________________________
            num += 1
        if num % DEBUG_NUM == 0:
            aaflog.debug("%d etablissements lus..." % num)

    db.commit('etablissement', no_bulk=no_bulk)
    aaflog.info("sous total : %d etablissements" % num)
    return num, etablissements_id

def delete_etab_educnat(ids):
    db_etabs = set()
    for etab in db.fetchall('etablissement', projection={'id':1}):
        db_etabs.add(etab['id'])
    etab_to_be_deleted = db_etabs - set(ids)
    for etab in etab_to_be_deleted:
        db_etab = db.fetchone('etablissement',
                              "id",
                              etab,
                              )
        modif_fields(db_etab, sqlgen, 'DELETE', modif_type='DELTADELETE', tablename="etablissement",
                     fromxml=True, db=db, no_bulk=True)
    db.commit('etablissement', no_bulk=True)


def parse_aaf_subjects(subject_file, allow_bulk, modif_type='CREATE', parse_type='ADD'):
    """Parsing XML des matières depuis AAF
    :etab_file: file name like ``*_MatiereEducNat_*.xml``
    """
    # init
    if allow_bulk and modif_type == 'CREATE' and not maj_mode:
        #cas aaf-complet (non maj)
        no_bulk = False
    else:
        no_bulk = True
    num = 0
    aaflog.info("Lecture des matieres...")
    subject_ids = []
    if parse_type == 'ADD':
        parse_tag = 'addRequest'
        parse_find = 'attributes/attr'

    elif parse_type == 'MODIFY':
        parse_tag = 'modifyRequest'
        parse_find = 'modifications/modification'

    action = _modif_type_to_action(modif_type)

    context = etree.iterparse(subject_file, events=('end',), tag=parse_tag)
    # ____________________________________________________________
    # dict construction by XML parsing
    for event, tresp in context:
        subject = dict()
        _type = tresp.find('operationalAttributes/attr/value').text
        if _type != 'MatEducNat':
            aaflog.info("%s n'est pas une matiere :(" % _type)
            continue
        subjectid = tresp.find('identifier/id').text
        if sys.version_info[0] < 3:
            subject['id'] = unicode(subjectid)
        else:
            subject['id'] = str(subjectid)
        subject_ids.append(subject['id'])
        for attr in tresp.findall(parse_find):
            balise, vals = parse_multi(attr)
            subject[balise] = vals
        retrieved = False
        launch_modif_fields = True
        if maj_mode or parse_type == 'MODIFY':
            db_subject = db.fetchone('group',
                                     "id",
                                     subject['id'],
                                     )
            if db_subject is None:
                modif_type = 'DELTACREATE'
                if parse_type == 'MODIFY':
                    aaflog.warn("ATTENTION: la matière {0} censée être mise a jour "
                             "n'existe pas dans la base".format(subject['id']))
                action = 'ADD'
            else:
                current_modif_type = db_subject['FieldActionType']
                if current_modif_type in ['CREATE', 'DELTACREATE']:
                    modif_type = 'DELTACREATE'
                elif current_modif_type == 'DELTADELETE':
                    modif_type = 'DELTADELETE'
                retrieved = True
                action = 'UPDATE'
        if retrieved:
            unlisted_key = set([u'FieldFromXMLImportActionDate', u'_id', u'FieldModifiedStatus', u'FieldActionType', u'FieldToLDIFFExportActionDate'])
            key_old = set(db_subject.keys())
            key_new = set(subject.keys())
            is_modified = False
            if len(key_old - key_new - unlisted_key) != 0 or len(key_new - key_old) != 0:
                is_modified = True
                aaflog.debug("some fields in the database ({}) and the current subject ({}) does not match for the subject: {}".format(
                          key_old - unlisted_key, key_new, subject['id']))
            else:
                for key in key_new:
                    if db_subject[key] != subject[key]:
                        is_modified = True
                        aaflog.debug("some entries in the database ({}) and the current subject ({}) does not match for the subject".format(
                                  db_subject[key], subject[key], subject['id']))
                        break
            if is_modified:
                if db_subject[u'FieldActionType'] in ['CREATE', 'DELTACREATE']:
                    modif_type = 'DELTACREATE'
                else:
                    modif_type = 'DELTAUPDATE'
                action = 'UPDATE'
            else:
                launch_modif_fields = False
        if launch_modif_fields:
            # database insert
            modif_fields(subject, sqlgen, action, modif_type=modif_type, tablename="subject",
                        fromxml=True, db=db, no_bulk=no_bulk)
            ## ____________________________________________________________
            num += 1
        if num % DEBUG_NUM == 0:
            aaflog.debug("%d matières lues..." % num)

    db.commit('group', no_bulk=no_bulk)
    aaflog.info("sous total: %d matières" % num)
    return num, subject_ids

def delete_subject(ids):
    db_subjects = set()
    for subject in db.fetchall('group', projection={'id':1}):
        db_subjects.add(subject['id'])
    subject_to_be_deleted = db_subjects - set(ids)
    for subject in subject_to_be_deleted:
        db_subject = db.fetchone('group',
                                 "id",
                                 subject,
                                 )
        modif_fields(db_subject, sqlgen, 'DELETE', modif_type='DELTADELETE', tablename="subject",
                     fromxml=True, db=db, no_bulk=True)
    db.commit('group', no_bulk=True)


# ____________________________________________________________
def reset_db():
    aaflog.info('Suppression de toutes les entrées de la base de données')
    db.reset()
    db.commit(None)
