# coding: utf-8
"""
Translator for samba
"""
import sys
from eoleaaf.config import basedn, smbstructure, loginshell, homedrive, domainnames, commonnames, generic_groups, toplevel_ou_user, toplevel_ou_group
from optparse import OptionParser
from samba.samdb import SamDB
from samba.auth import system_session
import samba.getopt as options
import ldb
import time
from eoleaaf.util import SimpleMongoConnectionWrapper, MongoDBGenerator, modif_fields
from eoleaaf.ihm import print_time


DEBUG_NUM = 100
INTERNAL_OBJ = ('CN=Builtin',
                'CN=Denied RODC Password Replication Group',
                'OU=Domain Controllers',
                'CN=admin',
                'CN=Guest',
                'CN=krbtgt',
                'CN=DnsUpdateProxy',
                'CN=Cert Publishers',
                'CN=RAS and IAS Servers',
                'CN=Domain Users',
                'CN=Read-only Domain Controllers',
                'CN=Domain Admins',
                'CN=Group Policy Creator Owners',
                'CN=Administrator',
                'CN=Domain Controllers',
                'CN=Domain Computers',
                'CN=Enterprise Admins',
                'CN=Enterprise Read-only Domain Controllers',
                'CN=Allowed RODC Password Replication Group',
                'CN=Schema Admins',
                'CN=Domain Guests',
                'CN=DnsAdmins',
                )
NOT_FOUND_ETABS = []


def not_internal(dn):
    """
    Verify if the entry (dn) can be deleted
    """
    first_level, second_level, _ = dn.split(',', 2)
    return not(first_level in INTERNAL_OBJ or second_level in INTERNAL_OBJ)


def add_administratif(ldbmessage):
    # it's hard to add objectClass during transaction, so do it here
    ldbmessage['objectClass'] = [ldbmessage['objectClass'],
                                 'posixAccount',
                                 'shadowAccount',
                                 'inetOrgPerson',
                                 'sambaSamAccount',
                                 #
                                 'administratif',
                                 'ENTPerson',
                                 'ENTAuxNonEnsEtab',
#                                 'radiusprofile',
                                 ]
    return ORI_ADD(ldbmessage)


def add_teacher(ldbmessage):
    # it's hard to add objectClass during transaction, so do it here
    ldbmessage['objectClass'] = [ldbmessage['objectClass'],
                                 'posixAccount',
                                 'shadowAccount',
                                 'inetOrgPerson',
                                 'sambaSamAccount',
                                 #
                                 'administrateur',
                                 'ENTPerson',
                                 'ENTAuxEnseignant',
#                                 'radiusprofile',
                                 ]
    return ORI_ADD(ldbmessage)


def add_student(ldbmessage):
    # it's hard to add objectClass during transaction, so do it here
    ldbmessage['objectClass'] = [ldbmessage['objectClass'],
                                 'posixAccount',
                                 'shadowAccount',
                                 'inetOrgPerson',
                                 'sambaSamAccount',
                                 #
                                 'Eleves',
                                 'ENTPerson',
                                 'ENTEleve',
#                                 'radiusprofile',
                                 ]
    return ORI_ADD(ldbmessage)


def add_group_eole(ldbmessage):
    ldbmessage['objectClass'] = [ldbmessage['objectClass'],
                                 'posixGroup',
                                 'sambaGroupMapping',
                                 'eolegroupe',
                                 'ENTGroupe',
                                 ]
    uai, niveau, ldbmessage['type'] = ldbmessage['info']
    del ldbmessage['info']
    if uai:
        ldbmessage['rne'] = uai
    if niveau:
        ldbmessage['objectClass'].append('classe')
        ldbmessage['objectClass'].append('ENTClasse')
        ldbmessage['niveau'] = niveau

    return ORI_ADD(ldbmessage)


def erase_db():
    """**CAUTION** use with care : erases all the samba users.
    use it at your own risks...
    """
    parse = OptionParser()
    lp = options.SambaOptions(parse).get_loadparm()
    creds = options.CredentialsOptions(parse).get_credentials(lp)
    samdb = SamDB(session_info=system_session(), credentials=creds, lp=lp)
    dn = samdb.domain_dn()
    #samdb.erase_users_computers(dn)
    try:
        res = samdb.search(base=dn, scope=ldb.SCOPE_SUBTREE, attrs=[],
                  expression="(|(objectclass=sambaFileShare)(objectclass=group)(objectclass=user)(objectclass=computer)(objectCategory=CN=Organizational-Unit,CN=Schema,CN=Configuration,{}))".format(dn))
    except ldb.LdbError as errno:
        if errno == ldb.ERR_NO_SUCH_OBJECT:
            # Ignore no such object errors
            pass
        else:
            raise

    ous = []
    for msg in res:
        olddn = str(msg.dn)
        if not_internal(olddn):
            if olddn.startswith('OU='):
                ous.append(msg.dn)
            else:
                samdb.delete(msg.dn, ["relax:0"])
    # OU is remove only if not toplevel import
    # delete OU after other entries
    for oudn in ous:
        try:
            samdb.delete(oudn, ["relax:0"])
        except Exception as err:
            print('impossible de supprimer {}: {}'.format(oudn, err))


def verify_indexed():
    # only for tests
    parse = OptionParser()
    lp = options.SambaOptions(parse).get_loadparm()
    lp.set("dsdb:schema update allowed", "yes")
    creds = options.CredentialsOptions(parse).get_credentials(lp)
    samdb = SamDB(session_info=system_session(), credentials=creds, lp=lp)
    indexed = False
    for res in samdb.search(base="@INDEXLIST", scope=ldb.SCOPE_BASE, attrs=['@IDXATTR']):
        for i in res['@IDXATTR']:
            if i == 'info':
                indexed = True
    if not indexed:
        modified_attributes = """
dn: CN=Comment,CN=Schema,CN=Configuration,{}
changetype: modify
replace: searchFlags
searchFlags: 1""".format(samdb.domain_dn())
        samdb.modify_ldif(modified_attributes)


class SmbCancel(Exception):
    pass

class SmbImportError(Exception):
    pass


class SambaImport(object):
    def __init__(self, log, debug=False, enable_transaction=False):
        self.sqlgen = MongoDBGenerator()
        self._init_connection(enable_transaction)
        self.log = log
        self.debug = debug
        self.invalid_classes = set()
        self.cache_groups = {}
        self.cache_etab_uai = []

    def _init_connection(self, enable_transaction):
        self.db = SimpleMongoConnectionWrapper()
        self.db.connect()
        parse = OptionParser()
        lp = options.SambaOptions(parse).get_loadparm()
        creds = options.CredentialsOptions(parse).get_credentials(lp)
        self.samdb = SamDB(session_info=system_session(), credentials=creds, lp=lp)
        global ORI_ADD
        ORI_ADD = self.samdb.add

        def fake_start():
            pass

        def fake_commit():
            pass

        def fake_cancel():
            self.transaction_cancel()
            raise SmbCancel("transaction cancelled")
        self.enable_transaction = enable_transaction
        self.transaction_start = self.samdb.transaction_start
        self.transaction_commit = self.samdb.transaction_commit
        self.transaction_cancel = self.samdb.transaction_cancel

        self.samdb.transaction_start = fake_start
        self.samdb.transaction_commit = fake_commit
        self.samdb.transaction_cancel = fake_cancel

    def get_typetab_name(self,
                         typ,
                         etab_name,
                         ):
        return etab_name + '_' + self.convert_groupname(typ)

    def get_group_dn(self, group, calculate=True):
        groupname = group['name']
        if groupname not in self.cache_groups:
            obj = self.samdb.search(basedn,
                                    scope=ldb.SCOPE_SUBTREE,
                                    expression="(&(objectClass=group)(cn={}))".format(groupname),
                                    attrs=["dn"],
                                    )
            if obj:
                if len(obj) > 1:
                    raise Exception(f'plusieurs groupes ont le nom {groupname}')
                self.cache_groups[groupname] = str(obj[0].dn)
            else:
                if calculate:
                    self.cache_groups[groupname] = self.calc_group_dn(group)
                else:
                    return None
        return self.cache_groups[groupname]

    def is_present_in_samba(self,
                            groupname,
                            etab_uai,
                            ):
        if etab_uai not in self.cache_etab_uai:
            domain_dn = self.calc_group_root_dn(etab_uai)
            ou, root_dn = domain_dn.split(',', 1)
            res = self.samdb.search(base=root_dn, expression=f'({ou})',
                                    scope=ldb.SCOPE_ONELEVEL)
            try:
                res[0].dn
            except IndexError:
                self.samdb.create_ou(domain_dn)
            self.cache_etab_uai.append(etab_uai)
        if self.get_group_dn({'name': groupname, 'etablissement_uai': etab_uai}, calculate=False):
            return True
        return False
#        except ldb.LdbError as err:
#            if err.args[0] == ldb.ERR_NO_SUCH_OBJECT:
#                # parent is not presente, create it
#                return False
#            raise err

    def share_present_in_samba(self,
                               sharename,
                               etab_uai,
                               ):
        domain_dn = self.calc_share_ou(etab_uai)
        if self.samdb.search(domain_dn, scope=ldb.SCOPE_ONELEVEL,
                    expression="(&(objectClass=sambaFileShare)(sambaShareName={}))".format(sharename),
                    attrs=["cn"]):
            return True
        return False

    def _create_or_modify_group(self,
                                group,
                                field,
                                etab_uai,
                                shared,
                                ):
        type_, export_type, row = field
        if export_type in ['CREATE', 'DELTACREATE']:
            self.log.debug('create new group {}'.format(group['groupname']))
            self._create_group(group,
                               etab_uai,
                               shared,
                               )
        elif export_type in ['DELTAUPDATE', 'DELTAMOVE']:
            if self.enable_transaction:
                raise SmbImportError('cannot update data with an activated transaction')
            self.log.debug('modifying group {}'.format(group['groupname']))
            group_dn = self.calc_group_dn({'name': group['groupname'], 'etablissement_uai': etab_uai})
            modified_attributes = """
dn: {0}
changetype: modify
replace: description
description: {1}
""".format(group_dn, group['description'])
            self.samdb.modify_ldif(modified_attributes)

    def _create_group(self,
                      group,
                      etab_uai,
                      shared,
                      ):
        self.log.debug('create new group {}'.format(group['groupname']))
        self.samdb.add = add_group_eole
        self.samdb.newgroup(**group)
        self.samdb.add = ORI_ADD
        self.cache_groups[group['groupname']] = self.calc_group_dn({'name': group['groupname'], 'etablissement_uai': etab_uai})
        if shared:
            groupname = group['groupname']
            model = None
            filepath = '/home/workgroups/{0}'.format(groupname)
            drive = None
            self.add_share(groupname,
                           groupname,
                           etab_uai,
                           model,
                           filepath,
                           drive,
                           )

    def _tag_create_or_modified_as_done(self, group):
        tablename, export_type, row = group
        if row:
            modif_export_type = export_type + "_DONE"
            modif_fields(row, self.sqlgen, action=None, modif_type=modif_export_type, tablename=tablename,
                         fromxml=False, db=self.db, id_column='id', no_bulk=True)
            self.db.commit(tablename, no_bulk=True)

    def create_groups(self,
                      export_type,
                      groups_id,
                      etabs_objects,
                      ):
        self.log.info('ajout des groupes "{}"'.format(export_type.lower()))
        cnt = 0
        transaction_started = False
        # searching for **all** groups
        # informations retrieved for all kind of group
        groups = []
        generic_groups_keys = list(generic_groups.keys())
        generic_groups_keys.sort()
        if export_type == 'CREATE':
            for idx, typ in enumerate(generic_groups_keys):
                groups_id[typ] = 20001 + idx
                kwargs = self.calc_group_kwargs(None,
                                                typ,
                                                'base',
                                                typ,
                                                groups_id[typ],
                                                )
                one_not_exists = False
                if not self.is_present_in_samba(kwargs['groupname'], None):
                    one_not_exists = True
                    shared = kwargs['groupname'] != 'eleves'
                    groups.append({'kwargs': kwargs,
                                   'etab_uai': None,
                                   'fields': ('base', 'CREATE', None),
                                   'shared': shared,
                                   })
                if one_not_exists:
                    # create base group, so create "commun"
                    self.add_share('commun',
                                   'commun',
                                   None,
                                   'commun',
                                   '/home/workgroups/commun',
                                   None,
                                   )
        for etab in etabs_objects.values():
            if etab['FieldActionType'] != export_type:
                continue
            etab_uai = etab['ENTStructureUAI']
            etab_name = commonnames.get(etab_uai, etab_uai)
            for idx, group in enumerate(generic_groups_keys):
                groupname = self.get_typetab_name(group,
                                                  etab_name,
                                                  )
                kwargs = self.calc_group_kwargs(etab_uai,
                                                groupname,
                                                'etablissement',
                                                etab['ENTStructureJointure'],
                                                )
                one_not_exists = False
                if not self.is_present_in_samba(kwargs['groupname'], etab_uai):
                    one_not_exists = True
                    shared = group != 'eleves'
                    groups.append({'kwargs': kwargs,
                                   'etab_uai': etab_uai,
                                   'fields': ('base', etab['FieldActionType'], None),
                                   'shared': shared,
                                  })
#                if one_not_exists and not self.share_present_in_samba('{}_commun'.format(etab_name), None):
#                    # create base group, so create "commun"
#                    print(')))', '{}_commun'.format(etab_name))
#                    self.add_share('{}_commun'.format(etab_name),
#                                   '{}_commun'.format(etab_name),
#                                   etab_uai,
#                                   'commun',
#                                   '/home/workgroups/{}_commun'.format(etab_name),
#                                   None,
#                                   )
            kwargs = self.calc_group_kwargs(etab_uai,
                                            etab_name,
                                            etab['ENTStructureNomCourant'],
                                            etab['ENTStructureJointure'],
                                            )
            groups.append({'kwargs': kwargs,
                           'etab_uai': etab_uai,
                           'fields': ('etablissement', etab['FieldActionType'], etab)
                          })
        def add_group(grp,
                      typ,
                      imported_group,
                      ):
            groupname = grp['name']
            etab_uai = grp['etablissement_uai']
            niveau = grp.get('niveau')
            if groupname in imported_group:
                return True
            imported_group.append(groupname)
            kwargs = self.calc_group_kwargs(etab_uai,
                                            groupname,
                                            typ,
                                            groupname,
                                            niveau=niveau
                                            )
            if typ == 'classe' and not niveau:
                self.invalid_classes.add(self.calc_group_dn(grp))
                if export_type == 'CREATE':
                    self.log.error("le groupe de la classe {} n'est pas créé car pas lié à un niveau".format(groupname))
                return False
            # do not update classe, niveau, ..., just create it if not already exists
            if not self.is_present_in_samba(kwargs['groupname'], etab_uai):
                if typ in ['niveau', 'classe', 'equipe']:
                    shared = True
                else:
                    shared = False
                groups.append({'kwargs': kwargs,
                               'etab_uai': etab_uai,
                               'fields': (typ, 'CREATE', None),
                               'shared': shared,
                               })
            return True
        for typ in ('classe', 'option', 'niveau', 'matiere'):
            imported_group = []
            for row in self.db.query_groupname(typ):
                if row:
                    if not isinstance(row, list):
                        row = [row]
                    for grp in row:
                        added = add_group(grp,
                                          typ,
                                          imported_group,
                                          )
                        if typ in ['classe', 'option']:
                            prof_grp = grp.copy()
                            prof_grp['name'] = 'profs-' + grp['name']
                            if added:
                                add_group(prof_grp,
                                          'equipe',
                                          imported_group,
                                          )
                            else:
                                self.invalid_classes.add(self.calc_group_dn(prof_grp))
        # _____________________________________________________________________
        # creates the groups or modifies the informations about the groups
        for group in groups:
            cnt += 1
            if self.enable_transaction and transaction_started is False:
                self.transaction_start()
                transaction_started = True
            try:
                field = group['fields']
                self._create_or_modify_group(group['kwargs'],
                                             field,
                                             group['etab_uai'],
                                             group.get('shared'),
                                             )
                if not self.enable_transaction:
                    self._tag_create_or_modified_as_done(field)
            except SmbCancel as err:
                if self.debug:
                    self.log.debug(err, exc_info=True)
                if self.enable_transaction:
                    raise err
                else:
                    self.log.error('cannot create or modify the group {}: {}'.format(group['kwargs']['groupname'], str(err)))
                    continue
            except Exception as err:
                if self.debug:
                    self.log.debug(err, exc_info=True)
                if self.enable_transaction:
                    self.transaction_cancel()
                    raise err
                else:
                    self.log.error('cannot create or modify the group {}: {}'.format(group['kwargs']['groupname'], str(err)))
                    continue
            if cnt % DEBUG_NUM == 0:
                self.log.debug("{} {}s traités...".format(cnt, 'groupe'))

        if transaction_started:
            self.transaction_commit()
            for group in groups:
                if group['fields']:
                    self._tag_create_or_modified_as_done(group['fields'])
        self.log.info("TOTAL : {} {}s".format(cnt, 'groupe'))

    def add_share(self,
                  sharename,
                  groupname,
                  etab_uai,
                  model,
                  filepath,
                  drive,
                  ):
        if not model:
            model = 'standard'
        if etab_uai is None:
            tetab_uai = etab_uai
        else:
            tetab_uai = etab_uai.upper()
        smb_server = domainnames.get(tetab_uai)
        if not smb_server:
            if etab_uai not in NOT_FOUND_ETABS:
                self.log.error("impossible de trouver le nom de domaine de l'établisement {}".format(etab_uai))
                NOT_FOUND_ETABS.append(etab_uai)
            return
        cn = "smb:{0}".format(sharename)
        ldbmessage = {"dn": "CN=" + cn + ',' + self.calc_share_ou(etab_uai),
                      "cn": cn,
                      "sambaShareName": sharename,
                      "description": sharename,
                      "sambaShareGroup": groupname,
                      "sambaFilePath": filepath,
                      "sambaShareURI": "\\\\{0}\\{1}".format(smb_server, sharename),
                      "uNCName": "\\\\{0}\\{1}".format(smb_server, sharename),
                      "sambaShareModel": model,
                      "objectClass": ["volume", "sambaFileShare"],
                      }
        if drive is not None:
            ldbmessage["sambaShareDrive"] = drive
        self.samdb.add(ldbmessage)

    def run(self):
        times = [time.time()]
        groups_id  = {}
        etabs_objects = {}
        etabs = self.db.fetchtable('etablissement')
        for etab in etabs:
            etabs_objects[etab['_id']] = dict(etab)
        etabs.close()
        for export_type in ['CREATE', 'DELTACREATE', 'DELTAUPDATE', 'DELTAMOVE']:
            self.create_groups(export_type, groups_id, etabs_objects)
        for typ in ('eleve', 'enseignant', 'administratif'):
            for export_type in ['CREATE', 'DELTACREATE', 'DELTAUPDATE', 'DELTAMOVE']:
                if export_type == 'CREATE':
                    self.log.info('ajout des {}s'.format(typ))
                elif export_type == 'DELTACREATE':
                    self.log.info('ajout des {}s (delta)'.format(typ))
                elif export_type == 'DELTAUPDATE':
                    self.log.info('mise à jour des {}s'.format(typ))
                elif export_type == 'DELTAMOVE':
                    self.log.info('déplacement des {}s'.format(typ))
                self.import_users(typ, export_type, groups_id, etabs_objects)
            print_time(typ, times)

    def add_or_modify_user(self,
                           typ,
                           export_type,
                           dico,
                           user_dn,
                           etab_uai,
                           etabs_objects,
                           ):
        if export_type in ["CREATE", "DELTACREATE"]:
            self.log.debug('add new user {}'.format(user_dn))
            username = str(dico['ENTPersonLogin'])
            if typ == 'eleve':
                self.samdb.add = add_student
            elif typ in ['enseignant', 'admin']:
                self.samdb.add = add_teacher
            elif typ == 'administratif':
                self.samdb.add = add_administratif
            else:
                raise Exception('unknown type {}'.format(typ))
            self.samdb.newuser(**self.calc_user_kwargs(username,
                                                       dico['userPassword'],
                                                       dico['gecos'],
                                                       etab_uai,
                                                       dico['givenName'],
                                                       dico['sn'],
                                                       dico['homedirectory'],
                                                       dico['profilepath'],
                                                       ))
            self.samdb.add = ORI_ADD
            ldif = """dn: {user_dn}
changeType: modify
add: uid
uid: {name}
add: ENTPersonLogin
ENTPersonLogin: {name}
""".format(user_dn=user_dn,
           name=username,
           )
            if 'personalTitle' in dico:
                ldif += """add: personalTitle
personalTitle: {title}
""".format(title=dico['personalTitle'])
            if typ == 'enseignant':
                if 'ENTAuxEnsClassesPrincipal' in dico and dico['ENTAuxEnsClassesPrincipal']:
                    ldif +="""add: Divcod
Divcod: {}
""".format(dico['ENTAuxEnsClassesPrincipal']['name'])
                    typeadmin = 2
                else:
                    typeadmin = 0
                ldif += """add: typeadmin
typeadmin: {}
""".format(typeadmin)
            elif typ == 'administratif':
                ldif += """add: typeadmin
typeadmin: 3
"""
            elif typ == 'eleve':
                ldif +="""add: Divcod
Divcod: {}
add: Meflcf
Meflcf: {}
""".format(dico['ENTEleveClasses']['name'],
           dico['ENTEleveNivFormation']['name'])
            #attrs.append((MOD_REPLACE, 'ENTPersonSexe', sexe))
            #datas.append((MOD_REPLACE, 'Divcod', args['classe']))
            self.samdb.modify_ldif(ldif)
            self.samdb.setexpiry("(&(objectClass=user)(sAMAccountName={}))".format(username), 2678400000, no_expiry_req=None)

            registered_groups = frozenset()
        elif export_type in ['DELTAUPDATE', 'DELTAMOVE']:
            self.log.debug('modify user {}'.format(user_dn))
            if export_type == 'DELTAMOVE':
                old_etab = etabs_objects.get(dico['ENTPersonStructRattachsOld'][-1])
                if old_etab:
                    old_etab_name = old_etab['ENTStructureUAI']
                    old_user_dn = self.calc_user_dn(str(dico['ENTPersonLogin']), old_etab_name)
                    self.move(old_user_dn, user_dn)
                else:
                    self.log.error("ne peut deplacer l'utilisateur {} dans le nouvel etablissement".format(user_dn))
            modified_attributes = """
dn: {0}
changetype: modify
replace: givenName
givenName: {1}
-
replace: displayName
displayName: {2}
-
replace: gecos
gecos: {2}
""".format(user_dn,
           dico['givenName'],
           dico['gecos'],
           )
            if dico['homedirectory']:
                modified_attributes += """-
replace: homeDirectory
homeDirectory: {0}
-
replace: profilePath
profilePath: {1}
""".format(dico['homedirectory'],
           dico['profilepath'],
           )
            self.samdb.modify_ldif(modified_attributes)
            # **registered_groups**, that is the groups where the user is already in
            registered_groups = set(self.get_groups_by_user(user_dn, etab_uai))
        return registered_groups

    def import_users(self,
                     typ,
                     export_type,
                     groups_id,
                     etabs_objects,
                     ):
        def add_group_etab(etab_uai, expected_groups, groups_infos):
            etab_name = commonnames.get(etab_uai, etab_uai)
            group = {'etablissement_uai': etab_uai,
                     'name': etab_name,
                     }
            group_dn = self.get_group_dn(group)
            expected_groups.add(group_dn)
            groups_infos[group_dn] = (etab_name, etab_uai)
            #
            group_name = self.get_typetab_name(typ, etab_name)
            group = {'etablissement_uai': etab_uai,
                     'name': group_name,
                     }
            group_dn = self.get_group_dn(group)
            groups_infos[group_dn] = (group_name, etab_uai)
            expected_groups.add(group_dn)

        cnt = 0
        transaction_started = False
        users_objects = []
        users = self.db.fetchtable(typ, export_type)
        for dico in users:
            users_objects.append(dico)
        users.close()
        for dico in users_objects:
            try:
                cnt += 1
                # **expected_groups** means the user is expected to belong to these groups
                group_name = self.convert_groupname(typ)
                group_dn = self.get_group_dn({'name': group_name, 'etablissement_uai': None})
                expected_groups = {group_dn}
                groups_infos = {group_dn: (self.convert_groupname(typ), None)}
                etab = etabs_objects.get(dico['ENTPersonStructRattach'])
                if etab:
                    etab_uai = etab['ENTStructureUAI']
                    add_group_etab(etab_uai, expected_groups, groups_infos)
                else:
                    etab_uai = None
                    etab_name = None
                user_dn = self.calc_user_dn(str(dico['ENTPersonLogin']), etab_uai)
                if transaction_started is False:
                    self.transaction_start()
                    transaction_started = True
                registered_groups = self.add_or_modify_user(typ,
                                                            export_type,
                                                            dico,
                                                            user_dn,
                                                            etab_uai,
                                                            etabs_objects,
                                                            )
                # let's add the etablissement at the user's group list
                # if it's a student, we add the student in his classroom
                if typ == 'eleve':
                    group_name = dico['ENTEleveClasses']['name']
                    group = {'etablissement_uai': etab_uai,
                             'name': group_name,
                             }
                    group_dn = self.get_group_dn(group)
                    groups_infos[group_dn] = (group_name, etab_uai)
                    expected_groups.add(group_dn)

                    # option
                    options = dico['ENTEleveGroupes']
                    if not isinstance(options, list):
                        options = [options]
                    for grp in options:
                        group_name = grp['name']
                        group_dn = self.get_group_dn(grp)
                        groups_infos[group_dn] = (group_name, etab_uai)
                        expected_groups.add(group_dn)
                elif typ == 'enseignant':
                    # matiere
                    matieres = dico['ENTAuxEnsMatiereEnseignEtab']
                    if not isinstance(matieres, list):
                        matieres = [matieres]
                    for grp in matieres:
                        group_name = grp['name']
                        group_dn = self.get_group_dn(grp)
                        groups_infos[group_dn] = (group_name, grp['etablissement_uai'])
                        expected_groups.add(group_dn)
                    # classe
                    etabs = [etab_uai]
                    classes = dico['ENTAuxEnsClasses']
                    if not isinstance(classes, list):
                        classes = [classes]
                    for grp in classes:
                        group = grp.copy()
                        group_name = 'profs-' + grp['name']
                        group['name'] = group_name
                        group_dn = self.get_group_dn(group)
                        groups_infos[group_dn] = (group_name, grp['etablissement_uai'])
                        expected_groups.add(group_dn)
                        if grp['etablissement_uai'] not in etabs:
                            # ajoute les etablissements autre que celui de rattachement
                            add_group_etab(grp['etablissement_uai'], expected_groups, groups_infos)
                            etabs.append(grp['etablissement_uai'])
                    # option
                    options = dico['ENTAuxEnsGroupes']
                    if not isinstance(options, list):
                        options = [options]
                    for grp in options:
                        group = grp.copy()
                        group_name = 'profs-' + grp['name']
                        group['name'] = group_name
                        group_dn = self.get_group_dn(group)
                        groups_infos[group_dn] = (group_name, grp['etablissement_uai'])
                        expected_groups.add(group_dn)
                elif typ == 'administratif':
                    groups = dico['ENTPersonFonctions']
                    if not isinstance(groups, list):
                        groups = [groups]
                    etabs = []
                    for grp in groups:
                        if grp['etablissement_uai'] not in etabs:
                            add_group_etab(grp['etablissement_uai'], expected_groups, groups_infos)
                            etabs.append(grp['etablissement_uai'])
#                        group_name = grp['name']
#                        group_dn = self.calc_group_dn(group_name, grp['etablissement_uai'])
#                        groups_infos[group_dn] = (group_name, grp['etablissement_uai'])
#                        expected_groups.add(group_dn)
                # let's add the user to the groups in which he is not already registered
                groups_to_add = expected_groups - registered_groups
                groups_to_add = groups_to_add - self.invalid_classes
                for group_dn in groups_to_add:
                    group_name, group_etab_uai = groups_infos[group_dn]
#                    if not self.is_present_in_samba(group_name, group_etab_uai):
#                        self.log.error("impossible d'inscrire {} dans le groupe {} dans l'établissement {} (c'est généralement une {} sans élève)".format(str(dico['ENTPersonLogin']), group_name, group_etab_uai, typ))
#                    else:
                    self.add_member_to_group(group_dn, user_dn)
                # let's remove the user to the groups in which he is not supposed to be registered
                groups_to_del = registered_groups - expected_groups
                for group in groups_to_del:
                    self.del_member_to_group(group, user_dn)
                # if the transaction_mode is not enabled,
                # we have to set the add/modify of the user as done
                if not self.enable_transaction:
                    modif_export_type = export_type + "_DONE"
                    modif_fields(dico, self.sqlgen, action=None, modif_type=modif_export_type, tablename=typ,
                                fromxml=False, db=self.db, id_column='id', no_bulk=True)
                    self.db.commit('user', no_bulk=True)
                if not self.enable_transaction:
                    self.transaction_commit()
                    transaction_started = False
                if cnt % DEBUG_NUM == 0:
                    self.log.info("{} {}s traités...".format(cnt, typ))
            except SmbCancel as err:
                transaction_started = False
                if self.debug:
                    self.log.debug(err, exc_info=True)
                if self.enable_transaction:
                    raise err
                else:
                    self.log.error('cannot create or modify the user {}: {}'.format(dico['ENTPersonLogin'], str(err)))
                    continue
            except Exception as err:
                if self.debug:
                    self.log.debug(err, exc_info=True)
                if transaction_started:
                    self.transaction_cancel()
                    transaction_started = False
                if self.enable_transaction:
                    raise err
                else:
                    self.log.error('cannot create or modify the user {}: {}'.format(dico['ENTPersonLogin'], str(err)))
                    continue
        if transaction_started:
            self.transaction_commit()
            cursor = self.db.fetchtable(typ, export_type)
            for row in cursor:
                modif_export_type = export_type + "_DONE"
                modif_fields(row, self.sqlgen, action=None, modif_type=modif_export_type, tablename=typ,
                            fromxml=False, db=self.db, id_column='id')
            cursor.close()
            self.db.commit('user')
        self.log.info("TOTAL : {} {}s".format(cnt, typ))

    def add_member_to_group(self,
                            group_dn,
                            member,
                            ):
        self.log.debug('add user {} to group {}'.format(member, group_dn))
        ldif = """dn: {}
changetype: modify
add: member
member: {}
""".format(group_dn, member)
        self.samdb.modify_ldif(ldif)

    def del_member_to_group(self, group_dn, member):
        self.log.debug('delete user {} to group {}'.format(member, group_dn))
        ldif = """dn: {}
changetype: modify
delete: member
member: {}
""".format(group_dn, member)
        self.samdb.modify_ldif(ldif)

    def get_groups_by_user(self,
                           username,
                           etab_uai,
                           ):
        user = username.split(',')[0]
        userfilter = "(&(%s)(objectCategory=%s,%s))" % (user, "CN=Person,CN=Schema,CN=Configuration", self.samdb.domain_dn())
        for res in self.samdb.search(base=self.calc_user_root_dn(etab_uai), scope=ldb.SCOPE_SUBTREE,
                expression=userfilter, attrs=['memberOf']):
            if 'memberOf' in res:
                for member in res['memberOf']:
                    if sys.version_info[0] < 3:
                        yield member
                    else:
                        yield member.decode()


class SambaTranslator(SambaImport):
    def convert_groupname(self, groupname):
        if groupname == 'eleve':
            return 'eleves'
        elif groupname == 'enseignant':
            return 'professeurs'
        elif groupname == 'administratif':
            return 'administratifs'
        return groupname

    def calc_user_kwargs(self,
                         username,
                         password,
                         gecos,
                         etab_uai,
                         givenname,
                         surname,
                         homedirectory,
                         profilepath):
        gecos = str(gecos.encode('utf-8'))
        return dict(username=username,
                    password=password,
                    userou=self.calc_user_ou(etab_uai),
                    givenname=givenname,
                    gecos=gecos,
                    surname=surname,
                    loginshell=loginshell,
                    profilepath=profilepath,
                    homedrive=homedrive,
                    homedirectory=homedirectory,
                    force_password_change_at_next_login_req=True,
                    useusernameascn=True)

    def calc_group_desc(self, grp, typ):
        if typ == 'classe':
            return 'Classe de {}'.format(grp)
        elif typ == 'option':
            return 'Option {}'.format(grp)
        elif typ == 'equipe':
            return 'Equipe {}'.format(grp)
        else:
            if sys.version_info[0] < 3:
                return 'Groupe {}'.format(typ.encode('utf-8'))
            else:
                return 'Groupe {}'.format(typ)

    def calc_group_kwargs(self,
                          etab_uai,
                          groupname,
                          typ,
                          ref,
                          niveau=None,
                          ):
        """group calculation

        :param etab_uai: name of the etablissement as created in the AD (usually UIA)
        :param groupname: name of the group as created in the AD
        :param typ: enables us tu build the group's description
        :param ref: id of the group in mongodb
        """
        if typ == 'classe':
            type_eole = (etab_uai, niveau, 'Classe')
        else:
            type_eole = (etab_uai, None, typ.capitalize())
        return dict(groupname=str(groupname),
                    groupou=self.calc_group_ou(etab_uai),
                    description=self.calc_group_desc(groupname, typ),
                    notes=type_eole,
                    )


# _____________________________________________________________________________
# Type of samba used (singleton determined by the configuration file)
class SambaToplevel(SambaTranslator):
    """Standard usage of the samba
    """
    def calc_group_ou(self, etab_uai):
        return toplevel_ou_group

    def calc_share_ou(self,
                      etab_uai,
                      ):
        return self.calc_group_root_dn(etab_uai)

    def calc_group_root_dn(self, etab_uai):
        return "{},{}".format(self.calc_group_ou(etab_uai), basedn)

    def calc_group_dn(self,
                      group,
                      ):
        groupname = self.convert_groupname(group['name'])
        etab_uai = group['etablissement_uai']
        return "CN={},{}".format(groupname, self.calc_group_root_dn(etab_uai))

    def calc_user_ou(self, etab_uai):
        return toplevel_ou_user

    def calc_user_root_dn(self, etab_uai):
        return "{},{}".format(self.calc_user_ou(etab_uai), basedn)

    def calc_user_dn(self, value, etab_uai):
        return "CN={},{}".format(value, self.calc_user_root_dn(etab_uai))

    def move(self, old_user_dn, user_dn):
        pass


class SambaGroupSplit(SambaToplevel):
    def calc_group_ou(self, etab_uai):
        if etab_uai is None:
            return toplevel_ou_group
        return f"OU={etab_uai},{toplevel_ou_group}"

    def calc_group_root_dn(self, etab_uai):
        return f"{self.calc_group_ou(etab_uai)},{basedn}"


class SambaSplit(SambaGroupSplit):
    """Splitted by OU (organisation unit)
    """
    def move(self, old_user_dn, user_dn):
        self.samdb.rename(old_user_dn, user_dn)

    def calc_group_ou(self, etab_uai):
        if etab_uai is None:
            return toplevel_ou_group
        return f"OU={etab_uai}"

    def calc_user_ou(self, etab_uai):
        if etab_uai is None:
            return toplevel_ou_user
        return f"OU={etab_uai}"


# _____________________________________________________________________________
# singleton definition
toplevel_factory = {
        'toplevel': SambaToplevel,
        'split': SambaSplit,
        'groupsplit': SambaGroupSplit,
        }

SambaExport = toplevel_factory[smbstructure]
# _____________________________________________________________________________
