# -*- coding: utf-8 -*-
###########################################################################
# EOLE - 2007-2020
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
###########################################################################
import sys
import time
from random import SystemRandom
from pymongo import MongoClient, InsertOne, UpdateOne

from pyeole.encode import normalize, remove_accents
from pyeole.log import init_logging

from .config import dbname, aaf_exceptions_report_filename, logfile, loglevel, \
                    ent_letter, ent_number, nb_generic_groups, \
                    remove_first_username_idx


FORCE_NO_BULK = False
#FORCE_NO_BULK = True


def format_name(name):
    return remove_accents(name).replace("'", "").replace(" ", "_").replace("-", "_").replace('&', '')


class DatabaseError(Exception):
    pass


random = SystemRandom()
# logging report utility
aaflog = init_logging(name="AAF", filename=logfile, level=loglevel)

# ____________________________________________________________
def log_error(tablename, modif_type, datestr, db, id_, msg):
    aaflog.error(msg)
    update_journal(tablename, modif_type, datestr, db, id_, 'error')


class SimpleMongoConnectionWrapper(object):
    def __init__(self):
        self._datas = []

    def connect(self):
        # FIXME: gérer l'authentification !!!!
        self.client = MongoClient('mongodb://localhost:27017')
        self.db = self.client[dbname]
        self.collection_user = self.db.user
        if "ENTEleveGroupes" not in self.collection_user.index_information():
            self.collection_user.create_index('ENTEleveGroupes')

        if "ENTPersonLogin" not in self.collection_user.index_information():
            self.collection_user.create_index('ENTPersonLogin', unique=True)

        if "ENTPersonJointure" not in self.collection_user.index_information():
            self.collection_user.create_index('ENTPersonJointure', unique=True)

        if "UserType" not in self.collection_user.index_information():
            self.collection_user.create_index('UserType')

        if "id" not in self.collection_user.index_information():
            self.collection_user.create_index('id')

        if "uid" not in self.collection_user.index_information():
            self.collection_user.create_index('uid')

        if "ENTStructureJointure" not in self.db.etablissement.index_information():
            self.db.etablissement.create_index('ENTStructureJointure', unique=True)

        if "ENTMatJointure" not in self.db.group.index_information():
            self.db.group.create_index('ENTMatJointure', unique=True)

    def reset(self):
        #self.client.drop_database(dbname)
        self.collection_user.remove()
        self.db.etablissement.remove()
        self.db.group.remove()
        self.db.setting.remove()
        self._datas = []

    def commit(self, tablename, no_bulk=False):
        if tablename is not None and self._datas:
            if not FORCE_NO_BULK and not no_bulk:
                self.db[tablename].bulk_write(self._datas)
            else:
                for data in self._datas:
                    try:
                        self.db[tablename].update_one(data[0], data[1], upsert=True)
                    except Exception as err:
                        aaflog.error('error when insert {} in {}: {}'.format(data[0], tablename, err))
        self._datas = []

    def close(self):
        self._datas = []
        self.client.close()

    def execute(self, data, values=None):
        self._datas.append(data)

    def fetchone(self,
                 table,
                 key,
                 value,
                 aaftype=None
                 ):
        filter_ = {key: value}
        if aaftype:
            filter_['UserType'] = aaftype
        return self.db[table].find_one(filter_)

    def fetchall(self,
                 tablename,
                 attr=None,
                 value=None,
                 dbfilter=None,
                 projection=None,
                 ):
        if dbfilter is not None:
            return self.db[tablename].find(dbfilter, no_cursor_timeout=True)
        elif attr is not None:
            return self.db[tablename].find({attr: value}, no_cursor_timeout=True)
        else:
            return self.db[tablename].find(projection=projection, no_cursor_timeout=True)

    def fetchtable(self,
                   tablename,
                   status=None,
                   ):
        """
        retrieves all entries from the table: tablename
        the same as fetchall without condition
        """
        attr = value = None
        dbfilter = None
        if tablename in ['eleve', 'enseignant', 'administratif', 'responsable']:
            dbfilter = {'UserType': tablename}
            tablename = 'user'
        if status:
            if dbfilter is None:
                dbfilter = {}
            dbfilter['FieldActionType'] = status

        return self.fetchall(tablename, dbfilter=dbfilter)

    def query_groupname(self, typ):
        """
        db.user.aggregate([{'$match':{"UserType":"eleve"}}, {'$group':{'_id': "$ENTEleveClasses.name"}}]);
        """
        query_id = None
        values = {}
        if typ == 'classe':
            query_id = "$ENTEleveClasses"
        elif typ == 'option':
            query_id = "$ENTEleveGroupes"
        elif typ == 'niveau':
            query_id = "$ENTEleveNivFormation"
        if query_id:
            for value in self.aggregate('eleve',
                                        query_id,
                                        ):
                if value['_id']:
                    if not isinstance(value['_id'], list):
                        val = [value['_id']]
                    else:
                        val = value['_id']
                    for v in val:
                        values[v['name']] = v
        if typ == 'classe':
            query_id = '$ENTAuxEnsClasses'
        elif typ == 'option':
            query_id = "$ENTAuxEnsGroupes"
        elif typ == 'matiere':
            query_id = '$ENTAuxEnsMatiereEnseignEtab'
        else:
            query_id = None
        if query_id:
            for value in self.aggregate('enseignant',
                                        query_id,
                                        ):
                if value['_id']:
                    if not isinstance(value['_id'], list):
                        val = [value['_id']]
                    else:
                        val = value['_id']
                    for v in val:
                        values[v['name']] = v
        return values.values()

    def aggregate(self,
                  user_type,
                  query_id,
                  ):
        dbfilter = {'$match': {'UserType': user_type}}
        return self.db.user.aggregate([dbfilter, {'$group': {'_id': query_id}}])

    def query_student_by_groupname(self, typ, groupname):
        """
        db.user.find({"UserType":"eleve", "ENTEleveClasses.name":"T STL"}, {ENTPersonLogin:1})
        """
        if typ == 'classe':
            query_id = "ENTEleveClasses.name"
        elif typ == 'groupe':
            query_id = "ENTEleveGroupes.name"
        return self.db.user.find({"UserType": "eleve", query_id: groupname}, {"ENTPersonLogin": 1}, no_cursor_timeout=True)


# ____________________________________________________________
# global (and unique) database connection
db = SimpleMongoConnectionWrapper()
db.connect()
# ____________________________________________________________

def replace_cars(chaine, chars):
    """
    fonction qui exécute le remplacement des caractères spéciaux
    """
    if chaine is None:
        return ''
    if sys.version_info[0] < 3 and isinstance(chaine, unicode):
         chaine = chaine.encode('UTF-8')
    for old, new in chars.items():
        chaine = chaine.replace(old, new)
    return chaine

def replace_cars2(chaine):
    """
    fonction de suppression des caractères spéciaux les plus courants
    """
    if chaine is None:
        return ''
    if sys.version_info[0] < 3 and isinstance(chaine, unicode):
         chaine = chaine.encode('UTF-8')
    chars = {'¥' : 'Y', 'µ' : 'u', 'À' : 'A', 'Á' : 'A',
             'Â' : 'A', 'Ã' : 'A', 'Ä' : 'A', 'Å' : 'A',
             'Æ' : 'A', 'Ç' : 'C', 'È' : 'E', 'É' : 'E',
             'Ê' : 'E', 'Ë' : 'E', 'Ì' : 'I', 'Í' : 'I',
             'Î' : 'I', 'Ï' : 'I', 'Ð' : 'D', 'Ñ' : 'N',
             'Ò' : 'O', 'Ó' : 'O', 'Ô' : 'O', 'Õ' : 'O',
             'Ö' : 'O', 'Ø' : 'O', 'Ù' : 'U', 'Ú' : 'U',
             'Û' : 'U', 'Ü' : 'U', 'Ý' : 'Y', 'ß' : 's',
             'à' : 'a', 'á' : 'a', 'â' : 'a', 'ã' : 'a',
             'ä' : 'a', 'å' : 'a', 'æ' : 'a', 'ç' : 'c',
             'è' : 'e', 'é' : 'e', 'ê' : 'e', 'ë' : 'e',
             'ì' : 'i', 'í' : 'i', 'î' : 'i', 'ï' : 'i',
             'ð' : 'o', 'ñ' : 'n', 'ò' : 'o', 'ó' : 'o',
             'ô' : 'o', 'õ' : 'o', 'ö' : 'o', 'ø' : 'o',
             'ù' : 'u', 'ú' : 'u', 'û' : 'u', 'ü' : 'u',
             'ý' : 'y', 'ÿ' : 'y', '~B': 'e', '°' : '.', 'A©' : 'e',
             'Š' : '', 'þ': 'e', '²': '',
#             '&#176;': '', '&#194;':'', '&#352;':'', '&#233;':'e',
#             '&#226;': 'a', '&#233;': 'e', '&#232;': 'e',
#             '&0xc3;': '', '&#231;': 'c', '&apos;': " ", '&#254;': 'e',
#              '&amp;': '', '&#224;':'', '&#201;': 'E', '&quot;': '',
#              '&#226;': 'a', '&#235;': 'e', "'": "", 'A©':'e',
#              '&#195;': 'a', '&#139;': '',
            }
    for old, new in chars.items():
        chaine = chaine.replace(old, new)
    return chaine

def replace_dangerous_cars(chaine):
    """
    remplace uniquement les caractères dangereux pour
    les traitements numériques conformément
    aux préconisations du SDET 4.2
    """
    chars = {'(': '', ')': '',
            '&quot;': "'",
            '&lt;': '<', # Less than
            '&gt;': '>', # Greater than
            '&amp;': "&", # eperluette
            '&apos;': "'", # apostrophe
#             '&#176;': '', '&#194;':'', '&#352;':'', '&#233;':'e',
#             '&#226;': 'a', '&#233;': 'e', '&#232;': 'e',
#             '&0xc3;': '', '&#231;': 'c', '&apos;': " ", '&#254;': 'e',
#             '&amp;': '', '&#224;':'', '&#201;': 'E', '&quot;': '',
#             '&#226;': 'a', '&#235;': 'e', "'": "", 'A©':'e',
#             '&#195;': 'a', '&#139;': '',
            }
    return replace_cars(chaine, chars)

def replace_bad_cars(chaine):
    """
    remplace les caractères invalides
    """
    longchars = {
            '\xc3\x82\xc2\xb0':'O',
            '\xc3\x83\xc2\x82':'A',
            '\xc3\x83\xc2\x87':'C',
            '\xc3\x83\xc2\x88':'E',
            '\xc3\x83\xc2\x89':'E',
            '\xc3\x83\xc2\x8a':'E',
            '\xc3\x83\xc2\x8b':'A',
            '\xc3\x83\xc2\x91':'N',
            '\xc3\x83\xc2\x94':'O',
            }
    shortchars = {
            '\xc3\xa8' : 'E',
            '\xc3\xaf' : 'I',
            '\xc3\x82' : 'A',
            '\xc3\x83' : 'E',
            '\xc3\x87' : '?',
            '\xc3\x88' : 'E',
            '\xc3\x89' : 'E',
            '\xc3\x8b' : 'E', #was ê maj
            '\xc3\x8f' : 'A',
            # ---------------- #
            '\xc3\x8a' : 'e', #was è
            '\xc3\xa7' : 'c', #was ç
            '\xc3\xa9' : 'e', #was é
            '\xc3\xaa' : 'e', #was ê
            '\xc3\xab' : 'e', #was ë
            '\xc3\xbb' : 'u',
            }

#             '&#176;': '', '&#194;':'', '&#352;':'', '&#233;':'e',
#             '&#226;': 'a', '&#233;': 'e', '&#232;': 'e',
#             '&0xc3;': '', '&#231;': 'c', '&#254;': 'e',
#             '&#224;':'', '&#201;': 'E',
#             '&#226;': 'a', '&#235;': 'e',
#             '&#195;': 'a', '&#139;': '',
#             '\xc2\x88' : '',
#             '\xc2\x89' : '',
#             '\xc2\x8a' : 'I',
#             '\xc2\x8b' : '',
#             '\xc2\xb0': '',
#             '\xc3\x82' : 'O',
    return replace_cars(replace_cars(chaine, longchars), shortchars)

def clean_empty_keys(empty_keys_dict):
    "delete empty keys to prepare a clen sql insert"
    for key, value in empty_keys_dict.items():
        if value in ['', None]:
            del(empty_keys_dict[key])
    return empty_keys_dict
## ______________________________________________________________________
## Mongo Engine data base ORM


class MongoDBGenerator(object):

    def insert(self, insert, table, paramstyle=False, no_bulk=False):
        # typing the collection
        if table not in ['etablissement', 'subject']:
            insert['UserType'] = table
        if not FORCE_NO_BULK and not no_bulk:
            return InsertOne(insert), None
        else:
            return self.update(insert, None, None, table, paramstyle, no_bulk=True)

    def update(self, notify, id_column, action_id, table, paramstyle=False, no_bulk=False):
        # typing the collection
        if table not in ['etablissement', 'subject']:
            notify['UserType'] = table
        filter_ = {'id': notify['id']}
        value = {'$set': notify}
        if not FORCE_NO_BULK and not no_bulk:
            return UpdateOne(filter_, value), None
        else:
            return (filter_, value), None


# ____________________________________________________________
def make_datetime():
    now = time.localtime(time.time())
    return time.strftime("%Y-%m-%d %H:%M:%S ", now)


def modif_fields(notify, sqlgen, action, modif_type,
                tablename, fromxml,
                db, id_column='id', no_bulk=False):
    """
    adds entries like update/modified tag and datetime

    :notifiy: dict representing the field to be notified
    :modif_type: in ["UPDATE", "CREATE", "DELETE", "LDIFEXPORT"]
    :fromxml: True si la modif est une importation xml,
              False sinon (la modif est une exportation ldif)
    """
    notify['FieldModifiedStatus'] = 1
    notify['FieldActionType'] = modif_type
    datestr = make_datetime()
    if fromxml:
        notify['FieldFromXMLImportActionDate'] = datestr
    else:
        notify['FieldToLDIFFExportActionDate'] = datestr
    action_id = notify[id_column]
    if action == 'ADD':
        #FIXME à voir si on fait un update ou on modifie directement
        if fromxml:
            sql, values = sqlgen.insert(notify, tablename, paramstyle=True, no_bulk=no_bulk)
        else:
            sql, values = sqlgen.update(notify, id_column, action_id, tablename, no_bulk=no_bulk)
    elif action == 'UPDATE':
        sql, values = sqlgen.update(notify, id_column, action_id, tablename,
                                    paramstyle=True, no_bulk=no_bulk)
    elif action == 'DELETE':
        if fromxml:
            sql, values = sqlgen.update(notify, id_column, action_id, tablename, no_bulk=no_bulk)
        else:
            raise NotImplementedError('not implemented!')
            ## suppression de l'entrée
            #journal = dict()
            #journal['ActionDate'] = datestr
            #journal['ActionId'] = action_id
            #journal['TableName'] = 'journal'
            #journal['Level'] = 'info'
            #journal['FieldActionType'] = modif_type
            #journal['FieldDeletedContent'] = str(notify).replace('"', '')
            ## FIXME : we commit in another table
            #sql = sqlgen.insert(journal, 'journal', no_bulk=no_bulk)
            #try:
            #    if paramstyle:
            #        db.execute(sql, values)
            #    else:
            #        db.execute(sql)
            #except Exception as msg:
            #    aaflog.info("Erreur au moment du storage du delete dans le journal : " + sql)
            #sql = sqlgen.delete(id_column, action_id, tablename)
    else:
        # on est ici seulement pour modifier le champ "FieldActionType"
        # action de log pour mette un _DONE
        #XXX ne changer que le status FieldActionType UNIQUEMENT !!!!!!!!!!!
        sql, values = sqlgen.update(notify, id_column, action_id, tablename,
                                    paramstyle=True, no_bulk=no_bulk)
    try:
        db.execute(sql)
    except Exception as msg:
        if "Duplicate entry" in str(msg):
            aaflog.info("Cette entrée est déjà présente dans la base de données au moment de {0} pour {1} avec ID {2}.".format(
                 modif_type, tablename, action_id.encode("utf-8")))
        else:
            aaflog.warn("Erreur au moment de {0} pour {1} avec ID {2} : {3}".format(
                       modif_type, tablename, action_id.encode("utf-8"), str(msg)))
    update_journal(tablename, modif_type, datestr, db, action_id, 'info')

# ____________________________________________________________
def update_journal(tablename, modif_type, datestr, db, id_, level):
    """
    every action shall add en entry in the journalisation table,
    in order to trace the actions in case of a failure

    :tablename: the table where the modification has been made
    :actiontype: in ['CREATE', 'UPDATE', 'DELETE']
    """
    return
    notify = dict()
    notify['ActionDate'] = datestr
    notify['ActionId'] = id_
    notify['TableName'] = tablename
    notify['Level'] = level
    notify['FieldActionType'] = modif_type

    # ecriture dans la table journal
    sql = sqlgen.insert(notify, 'journal')
    db.execute(sql)

#____________________________________________________________
# random generators

def random_letter(upper=False):
    "[A-Z]"
    letter = random.choice('abcdefghijklmnopqrstuvwxyz')
    return letter.upper() if upper else letter

def random_number():
    "[0-9]"
    return str(random.randint(0, 9))

def random_upper():
    "[A-Z]"
    return random_letter(upper=True)

def random_name(length=5):
    return "".join([random_upper() for i in range(length)])

def random_id(length=5):
    return "".join([random_number() for i in range(length)])

def random_special_cars():
    "our special chars"
    return random.choice(',;:&$@%=')

def random_identifier(nblength=4, namelength=2):
    rndid = random_id(length=nblength)
    rndname = random_name(length=namelength)
    return rndname + "-" + rndid

#____________________________________________________________
# identifier shall be unique in the whole file
#ids_buffer = []
#def read_id_file(filename):
#    global ids_buffer
#    if not isfile(filename):
#        return
#    fh = open(filename, "r")
#    for line in fh.readlines():
#        strid = line.strip()
#        ids_buffer.append(strid)
#    fh.close()


#def write_id_file(filename, newid):
#    """
#    add the new id in id_file
#    """
#    global ids_buffer
#    fh = open(filename, 'a+')
#    fh.write(newid + '\n')
#    ids_buffer = []
#    fh.close()


def make_unique_random_id(rnd_buffer):
    """generates a unique id for the 'ENTPersons'
    it uses the primary key in the 'randomids' sql table
    while an id exists in the table, it is calculated again
    """
    rnd = random_identifier()
    rndfind = True
    idx = 0
    while rndfind:
        idx += 1
        try:
            if rnd in rnd_buffer or db.fetchone("user",
                                                "uid",
                                                rnd,
                                                ):
                rnd = random_identifier()
            else:
                rndfind = False
                rnd_buffer.append(rnd)
        except DatabaseError:
            rnd = random_identifier()
        if idx == 10:
            raise Exception('cannot generate ID')

    return rnd.split("-")

#______________________________________________________________________________

def get_new_id(rnd_buffer):
    """
    Renvoie un identifiant unique
    de la forme "LxxCiiii" avec :
    o L et C : Lettre et chiffre du code projet ENT
    o xx et iiii : 2 lettres et 4 chiffres à générer pour chaque entrée
    """
    letters, numbers = make_unique_random_id(rnd_buffer)
    return "{0}{1}{2}{3}".format(ent_letter, letters, ent_number, numbers)


#______________________________________________________________________________
def gen_user_login(surname, givenname, uid, usertype, user_id, login_buffer):
    """
    génération du login
    le login est de type 'prenom.nomXX'
    """
    def gen_login_num(num):
        if remove_first_username_idx and num == 1:
            return ''
        if num < 9:
            return "0" + str(num)
        return str(num)

    givenname = normalize(givenname)
    surname = normalize(surname)
    # C'EST CONTRAIRE AU SDET - on cherche quand même a attribuer un login a l'utilisateur
    loginalpha = ""
    if surname is not None or surname != "":
        if (givenname is not None and givenname.strip() not in ['-', '_', '']
                and len(givenname + surname) <= 48):
                # Concatenation must be lower than 48 because we add 2 digits
            loginalpha = givenname.lower() + '.' + surname.lower()
        else:
            loginalpha = surname.lower()
        loginalpha = format_name(loginalpha)

    if loginalpha == "":
        # C'EST CONTRAIRE AU SDET - on cherche quand même a attribuer un login a l'utilisateur
        # on lui attribut un login correspondant au user id au sens AAF (pas l'uid généré)
        loginalpha = user_id
        write_report = True
    else:
        write_report = False
    rndfind = True
    idx = 0
    double = '__' in loginalpha
    while double:
        loginalpha = loginalpha.replace('__', '_')
        double = '__' in loginalpha
    while rndfind:
        try:
            idx += 1
            if idx == 101:
                raise Exception('cannot generate login')
            loginnum = gen_login_num(idx)
            login = loginalpha + loginnum
            if login in login_buffer or db.fetchone("user",
                                                    "ENTPersonLogin",
                                                    login,
                                                    ):
                continue
            login_buffer.append(login)
            rndfind = False
        except DatabaseError:
            pass
        except Exception as msg:
            raise Exception('Impossible de créer le login pour {0} : {1}'.format(loginalpha, msg))
    if write_report:
        # on ecrit un rapport sur les cas exceptionnels dans un fichier texte
        fh = file(aaf_exceptions_report_filename, 'a+')
        fh.write("l'utilisateur avec l'id: {0} se voit affecter le login : {1}\n".format(user_id, login))
        fh.close()
    return login

def parse_multi(attr):
    """parse des balises XML qui sont possiblement multi
    (plusieurs balises VALUE)

    :param attr: noeud XML
    """
    balise = attr.attrib['name']
    vals = []
    for val in attr.findall('value'):
        val = replace_dangerous_cars(val.text)
        if sys.version_info[0] < 3 and isinstance(val, unicode):
             val = val.decode('utf-8')
        vals.append(val)
    if len(vals) == 1:
        vals = vals[0]
    operation = attr.attrib.get('operation')
    if operation not in [None, "replace"]:
        aaflog.error("operation de type inconnu : {}".format(operation))
    return balise, vals

def pwgen():
    "convenient password generator"
    random_factory = {
        1: random_letter,
        2: random_upper,
        3: random_number,
        4: random_special_cars}
    # let's decide how to populate the passwd string
    # default weighting ([a-z] : 4, [A-Z] : 3, [0-9] : 2, other : 2)
    weighting = [1, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4]
    return "".join([random_factory[n]() for n in random.sample(weighting, 8)])
    #cmd = "pwgen -1 -c -n"


def gen_user_displayname(surname,
                         givenname,
                         ):
    """
    Génération de l'attribut `displayName`.
    Le `displayName` se construit de la manière suivante :
    Nom et prénom accentués

    :param surname: nom d'usage
    :param givenname: prénom usuel
    """
    if givenname:
        return givenname.title() + ' ' + surname.upper()
    return surname.upper()
